// Copyright 2011 Google Inc. All Rights Reserved.

/**
 * @fileoverview Highlights a findy tidbit results in a search document page.
 * @author jonwall@google.com (Jon Wall)
 */

/**
 * Namespace for this file.
 */
var ipadFindy = {};

/**
 * Highlighted tidbit results.
 * @type {ipadFindy.TidbitResult}
 */
ipadFindy.resultsInDocument = null;

/**
 * Tidbit result scraped from google search results page.
 * @type {Findy_Result}
 */
ipadFindy.scrapedResults = null;

/**
 * IFrame used for application callback.
 * @type {Element}
 */
ipadFindy.callbackFrame = null;

/** type {RegEx} */
ipadFindy.REGEX_NONSPACE = /\S/;

/** @type {RegExp} */
ipadFindy.REGEX_CSS_CLASS_NAME = new RegExp('\\s' + findInPage.CSS_CLASS_NAME);

/**
 * Initial function to kickoff the finding of tidbits. This call is
 * asynchronous; it calls ipadFindy.onFoundTidbits when complete.
 * Make sure the DOM is in a loaded state, otherwise add an event listener.
 */
ipadFindy.findTidbits = function() {
  if (ipadFindy.scrapedResults) {
    if (document.readyState == 'interactive' ||
        document.readyState == 'complete') {
      ipadFindy.doFindTidbits();
    } else {
      document.addEventListener('DOMContentLoaded', ipadFindy.doFindTidbits);
    }
  }
};

/**
 * Interface function to the findy code, called by ipadFindy.findTidbits when
 * the DOM is loaded.
 */
ipadFindy.doFindTidbits = function() {
  ipadFindy.initCallbackFrame();
  var findyHighlighter = new Findy_TidbitHighlighter(window,
                                                     document,
                                                     ipadFindy.onFoundTidbits);
  findyHighlighter.showTidbitsOnPage(ipadFindy.scrapedResults);
};

/**
 * Callback function for Findy_TidbitHighlighter.
 * @param {Findy_Result} result The original Findy_Result JSON object.
 * @param {Array.<Findy_FoundTidbit>} foundTidbits The tidbits found.
 */
ipadFindy.onFoundTidbits = function(result, foundTidbits) {
  if (foundTidbits && foundTidbits.length) {
    ipadFindy.resultsInDocument = new ipadFindy.TidbitResult(foundTidbits);
    if (ipadFindy.resultsInDocument.tidbits.length) {
      ipadFindy.appCallback();
    }
  }
};

// Also defined in GMOiPadWebPaneController.m.
ipadFindy.CALLBACK_URL = 'http://www.google.com/empty.html';
ipadFindy.CALLBACK_FRAGMENT = '#found_findy_tidbits';
ipadFindy.REGEX_CALLBACK_FRAGMENT = new RegExp(ipadFindy.CALLBACK_FRAGMENT);

ipadFindy.appCallback = function() {
  ipadFindy.callbackFrame.src += ipadFindy.CALLBACK_FRAGMENT;
};

ipadFindy.initCallbackFrame = function() {
  if (ipadFindy.callbackFrame) {
    ipadFindy.callbackFrame.src = ipadFindy.callbackFrame.src.replace(
        ipadFindy.REGEX_CALLBACK_FRAGMENT, '');
  }
  else {
    var iframe = document.createElement('iframe');
    iframe.name = 'gsa_findy_frame';
    iframe.style.display = 'none';
    iframe.src = ipadFindy.CALLBACK_URL;
    document.body.appendChild(iframe);
    ipadFindy.callbackFrame = iframe;
  }
};

/**
 * Adds the CSS styles to the page and adds the highlight css to the tidbits.
 * @param {number} width Width of display screen.
 * @param {number} height Height of display screen.
 */
ipadFindy.highlightTidbits = function(width, height) {
  if (ipadFindy.resultsInDocument) {
    findInPage.init(width, height);
    findInPage.enable();
    ipadFindy.resultsInDocument.addHighlighting();
  }
};

/**
 * Remove all findy highlighting.
 */
ipadFindy.removeHighlighting = function() {
  findInPage.disable();
  ipadFindy.resultsInDocument.removeHighlighting();
};

/**
 * Called from app when tidbit is selected.
 * @param {number} tidbitIndex Index of the selected tidbit.
 * @return {string} Position of the highlighted tidbit.
 */
ipadFindy.tidbitSelected = function(tidbitIndex) {
  var position = ipadFindy.resultsInDocument.tidbitSelected(tidbitIndex);
  return '[' + position.join(',') + ']';
};

/**
 * Retrieval method to get results back. Called from GMA app.
 * @return {string} JSON array of tidbit strings.
 */
ipadFindy.getHighlightedTidbits = function() {
  if (ipadFindy.resultsInDocument) {
    return ipadFindy.resultsInDocument.toJsonString();
  } else {
    return '[]';
  }
};

/**
 * Individual found tidbit result.
 * @param {Findy_FoundTidbit} foundTidbit The tidbit to wrap.
 * @constructor
 */
ipadFindy.Tidbit = function(foundTidbit) {
  /** type {string} */
  this.tidbit = foundTidbit.getTidbit();
  /** type {Array.<Element>} */
  this.elements = foundTidbit.getElements();
  this.wrapElementsWithSpan();
};

/**
 * Try to wrap all of the elements in this tidbit within a single span.
 */
ipadFindy.Tidbit.prototype.wrapElementsWithSpan = function() {
  if (!this.elements || this.elements.length < 2) {
    return;
  }
  var firstElement = this.elements[0];
  var lastElement = this.elements[this.elements.length - 1];

  var safeToWrap = false;
  if (firstElement.parentNode.isSameNode(lastElement.parentNode)) {
    // First and last elements are siblings. Now make sure none of the elements
    // are block.
    safeToWrap = true;
    var length = this.elements.length;
    for (var i = 0; i < length; i++) {
      var node = this.elements[i];
      if (this.isBlockElement(node)) {
        safeToWrap = false;
        break;
      }
    }
  } else {
    var elementsToWrap = this.findParentElementsToWrap();
    if (elementsToWrap) {
      // Walk up the tree to see if it's safe to wrap elements within a span.
      safeToWrap =
          this.isSafeToWrapAtParent(firstElement, elementsToWrap[0], true) &&
          this.isSafeToWrapAtParent(lastElement, elementsToWrap[1], false);
      if (safeToWrap) {
        firstElement = elementsToWrap[0];
        lastElement = elementsToWrap[1];
      }
    }
  }

  if (safeToWrap) {
    // Now do the actual wrapping.
    var container = document.createElement('span');
    firstElement.parentNode.insertBefore(container, firstElement);
    var current = firstElement;
    var endPoint = lastElement.nextSibling;
    while (current && !current.isSameNode(endPoint)) {
      var next = current.nextSibling;
      container.appendChild(current);
      current = next;
    }
    // Set the wrapper as our only element.
    this.elements = [container];
  }
};

/**
 * Checks if the specified element is a block element.
 * @param {Element} element Element to check.
 * @return {boolean} True if the element has display 'block' or 'inline-block'.
 */
ipadFindy.Tidbit.prototype.isBlockElement = function(element) {
  var display = window.getComputedStyle(element).display;
  return display == 'block' || display == 'inline-block';
};

/**
 * Finds the lowest common ancestor of this tidbit's elements and returns the
 * leftmost and rightmost ancestors directly below the LCA. Used to wrap the
 * elements in the tidbit at the lowest point possible.
 * @return {Array.<Element>} The left and rightmost uncommon ancestors.
 */
ipadFindy.Tidbit.prototype.findParentElementsToWrap = function() {
  var stacks = [];
  this.elements.forEach(function(element) {
    var stack = [];
    while (element) {
      stack.unshift(element);
      element = element.parentNode;
    }
    stacks.push(stack);
  });

  var numStacks = stacks.length;
  var stackLength = stacks[0].length;
  for (var i = 0; i < stackLength; i++) {
    var node = stacks[0][i];
    for (var j = 1; j < numStacks; j++) {
      if (!node.isSameNode(stacks[j][i])) {
        return [stacks[0][i], stacks[numStacks - 1][i]];
      }
    }
  }
  return null;
};

/**
 * Walk up the DOM from node to parentNode checking for two things:
 * 1. There are no block level elements between the two nodes.
 * 2. There is no text content in sibling nodes located 'outside' of the tidbit
 *    elements, but still inside the parentNode. For example, this is not
 *    wrappable:
 *    <div class=outer>
 *      <span>
 *        Outside text.
 *        <span class=highlight>
 *          Highlighted text.
 *        </span>
 *      </span>
 *      <span class=highlight>
 *        Highlighted text.
 *      </span>
 *    </div>
 * @param element Node to check.
 * @param parentNode Ancestor node to stop at.
 * @param isFirst If this is the first element in the tidbit block.
 * @return {boolean} True if the node is safe to wrap.
 */
ipadFindy.Tidbit.prototype.isSafeToWrapAtParent = function(element,
                                                           parentNode,
                                                           isFirst) {
  var getSibling = function(element) {
    return isFirst ? element.previousSibling : element.nextSibling;
  };
  var current = element;
  while (!current.isSameNode(parentNode.parentNode)) {
    if (!current.isSameNode(parentNode)) {
      // If this isn't the parentNode, check siblings for text content.
      var sibling = getSibling(current);
      while (sibling) {
        if ((sibling.nodeType == 1 || sibling.nodeType == 3) &&
            sibling.textContent &&
            ipadFindy.REGEX_NONSPACE.test(sibling.textContent)) {
          return false;
        }
        sibling = getSibling(sibling);
      }
    }
    if (this.isBlockElement(current)) {
      // We cannot wrap around a block level element.
      return false;
    }
    current = current.parentNode;
  }
  return true;
};

/**
 * Add the findy css class to this tidbit.
 */
ipadFindy.Tidbit.prototype.addClass_ = function() {
  this.elements.forEach(function(element) {
    element.className =
        (element.className || '') + ' ' + findInPage.CSS_CLASS_NAME;
  });
};

/**
 * Remove the findy css class from this tidbit.
 */
ipadFindy.Tidbit.prototype.removeClass_ = function() {
  this.elements.forEach(function(element) {
    element.className =
        element.className.replace(ipadFindy.REGEX_CSS_CLASS_NAME, '');
  });
};

/**
 * Class encapsulating a result set of found tidbits. This is a simplified
 * version of Findy_FoundTidbit that includes just the tidbit and position.
 * @param {Array.<Findy_FoundTidbit>} foundTidbits The tidbits to wrap.
 * @constructor
 */
ipadFindy.TidbitResult = function(foundTidbits) {
  var arr = [];
  foundTidbits.forEach(function(tidbit) {
    var elements = tidbit.getElements();
    if (elements && elements.length && findInPage.isVisible(elements[0])) {
      arr.push(new ipadFindy.Tidbit(tidbit));
    }
  });
  /** type {Array.<Tidbit>} */
  this.tidbits = arr;
  /** type {number} */
  this.selectedTidbitIndex = -1;
};

/**
 * Adds the highlighting styles to the tidbits.
 */
ipadFindy.TidbitResult.prototype.addHighlighting = function() {
  this.tidbits.forEach(function (tidbit) {
    tidbit.addClass_();
  });
};

/**
 * Remove all highlighting for this result.
 */
ipadFindy.TidbitResult.prototype.removeHighlighting = function() {
  this.deselectTidbit();
  this.tidbits.forEach(function (tidbit) {
    tidbit.removeClass_();
  });
};

/**
 * Tidbit at the specified index was selected.
 * @param {number} tidbitIndex Index of tidbit selected.
 * @return {Array.<number>} Position of selected tidbit.
 */
ipadFindy.TidbitResult.prototype.tidbitSelected = function(tidbitIndex) {
  this.deselectTidbit();
  this.selectedTidbitIndex = tidbitIndex;
  var tidbit = this.tidbits[tidbitIndex];
  tidbit.elements.forEach(findInPage.addSelectHighlight);
  return findInPage.scaleCoordinates(
      findInPage.getNormalizedCoordinates(tidbit.elements[0]));
};

/**
 * Remove the selected highlighting for the currently selected tidbit.
 */
ipadFindy.TidbitResult.prototype.deselectTidbit = function() {
  if (this.selectedTidbitIndex >= 0) {
    this.tidbits[this.selectedTidbitIndex].elements.forEach(
        findInPage.removeSelectHighlight);
  }
  this.selectedTidbitIndex = -1;
};

/**
 * Creates a JSON representation of the result object. It's risky to use
 * JSON.stringify() directly, as we are running this JavaScript on third-party
 * sites and some have libraries that change the default behavior.
 * @return {string} JSON representation of the result.
 */
ipadFindy.TidbitResult.prototype.toJsonString = function() {
  var tidbitStrings = [];
  this.tidbits.forEach(function(element) {
    tidbitStrings.push(JSON.stringify(element.tidbit));
  });
  return '[' + tidbitStrings.join(',') + ']';
};
